@@ -659,12 +659,91 @@ class ChemBFN(nn.Module):
         return (-logits.gather(-1, x[..., :1]).squeeze(-1)).mean()
 
     @staticmethod
-    def reshape_y(y: Tensor) -> Tensor:
+    def _reshape(y: Tensor) -> Tensor:
         assert y.dim() <= 3  # this doesn't work if the model is frezen in JIT.
         if y.dim() == 2:
             return y[:, None, :]
         return y
 
+    def _process(
+        self,
+        theta: Tensor,
+        mask: Optional[Tuple[Tensor, Tensor]],
+        y: Optional[Tensor],
+        sample_step: int,
+        guidance_strength: float,
+        token_mask: Optional[Tensor],
+    ) -> Tuple[Tensor, Tensor]:
+        # BFN inference process.
+        #
+        # theta: piror distribution;            shape: (n_b, n_t, n_vocab)
+        # mask: masked condition distribution;  shape: (n_b, n_t, n_vocab)
+        #       condition distribution mask;    shape: (n_b, n_t, 1)
+        n_b = theta.shape[0]
+        if y is not None:
+            y = self._reshape(y)
+        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
+            t = (i - 1).view(1, 1, 1).repeat(n_b, 1, 1) / sample_step
+            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
+            if token_mask is not None:
+                p = p.masked_fill_(token_mask, 0.0)
+            alpha = self.calc_discrete_alpha(t, t + 1 / sample_step)
+            e_k = nn.functional.one_hot(torch.argmax(p, -1), self.K).float()
+            mu = alpha * (self.K * e_k - 1)
+            sigma = (alpha * self.K).sqrt()
+            theta = (mu + sigma * torch.randn_like(mu)).exp() * theta
+            theta = theta / theta.sum(-1, True)
+            if mask is not None:
+                x_onehot, x_mask = mask
+                theta = x_onehot + (1 - x_mask) * theta
+        t_final = torch.ones((n_b, 1, 1), device=self.beta.device)
+        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
+        entropy = -(p * p.log()).sum(-1).mean(-1)
+        if token_mask is not None:
+            p = p.masked_fill_(token_mask, 0.0)
+        return torch.argmax(p, -1), entropy
+
+    def _ode_process(
+        self,
+        z: Tensor,
+        mask: Optional[Tuple[Tensor, Tensor]],
+        y: Optional[Tensor],
+        sample_step: int,
+        guidance_strength: float,
+        token_mask: Optional[Tensor],
+        temperature: float,
+    ) -> Tuple[Tensor, Tensor]:
+        # ODE-solver engaged inference process.
+        #
+        # z: prior latent vector;               shape: (n_b, n_t, n_vocab)
+        # mask: masked condition distribution;  shape: (n_b, n_t, n_vocab)
+        #       condition distribution mask;    shape: (n_b, n_t, 1)
+        n_b = z.shape[0]
+        if y is not None:
+            y = self._reshape(y)
+        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
+            t = (i - 1).view(1, 1, 1).repeat(n_b, 1, 1) / sample_step
+            theta = softmax(z, -1)
+            if mask is not None:
+                x_onehot, x_mask = mask
+                theta = x_onehot + (1 - x_mask) * theta
+            beta = self.calc_beta(t + 1 / sample_step)
+            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
+            if token_mask is not None:
+                p = p.masked_fill_(token_mask, 0.0)
+            u = torch.randn_like(z)
+            z = (self.K * p - 1) * beta + (self.K * beta * temperature).sqrt() * u
+        t_final = torch.ones((n_b, 1, 1), device=self.beta.device)
+        theta = softmax(z, -1)
+        if mask is not None:
+            x_onehot, x_mask = mask
+            theta = x_onehot + (1 - x_mask) * theta
+        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
+        entropy = -(p * p.log()).sum(-1).mean(-1)
+        if token_mask is not None:
+            p = p.masked_fill_(token_mask, 0.0)
+        return torch.argmax(p, -1), entropy
+
     @torch.jit.export
     def sample(
         self,
@@ -676,48 +755,30 @@ class ChemBFN(nn.Module):
         token_mask: Optional[Tensor] = None,
     ) -> Tuple[Tensor, Tensor]:
         """
-        Sample from a piror distribution.
+        Sample from a uniform piror distribution.
 
         :param batch_size: batch size
         :param sequence_size: max sequence length
-        :param y: conditioning vector;      shape: (n_b, 1, n_f) or (n_b, n_f)
+        :param y: conditioning vector;   shape: (n_b, 1, n_f) or (n_b, n_f)
         :param sample_step: number of sampling steps
         :param guidance_strength: strength of conditional generation. It is not used if y is null.
         :param token_mask: token mask assigning unwanted token(s) with `True`;
-                                            shape: (1, 1, n_vocab)
+                                         shape: (1, 1, n_vocab)
         :type batch_size: int
         :type sequence_size: int
         :type y: torch.Tensor | None
         :type sample_step: int
         :type guidance_strength: float
         :type token_mask: torch.Tensor | None
-        :return: sampled token indices;     shape: (n_b, n_t) \n
-                 entropy of the tokens;     shape: (n_b)
+        :return: sampled token indices;  shape: (n_b, n_t) \n
+                 entropy of the tokens;  shape: (n_b)
         :rtype: tuple
         """
         theta = (
             torch.ones((batch_size, sequence_size, self.K), device=self.beta.device)
             / self.K
         )
-        if y is not None:
-            y = self.reshape_y(y)
-        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
-            t = (i - 1).view(1, 1, 1).repeat(batch_size, 1, 1) / sample_step
-            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
-            if token_mask is not None:
-                p = p.masked_fill_(token_mask, 0.0)
-            alpha = self.calc_discrete_alpha(t, t + 1 / sample_step)
-            e_k = nn.functional.one_hot(torch.argmax(p, -1), self.K).float()
-            mu = alpha * (self.K * e_k - 1)
-            sigma = (alpha * self.K).sqrt()
-            theta = (mu + sigma * torch.randn_like(mu)).exp() * theta
-            theta = theta / theta.sum(-1, True)
-        t_final = torch.ones((batch_size, 1, 1), device=self.beta.device)
-        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
-        entropy = -(p * p.log()).sum(-1).mean(-1)
-        if token_mask is not None:
-            p = p.masked_fill_(token_mask, 0.0)
-        return torch.argmax(p, -1), entropy
+        return self._process(theta, None, y, sample_step, guidance_strength, token_mask)
 
     @torch.jit.export
     def ode_sample(
@@ -735,11 +796,11 @@ class ChemBFN(nn.Module):
 
         :param batch_size: batch size
         :param sequence_size: max sequence length
-        :param y: conditioning vector;      shape: (n_b, 1, n_f) or (n_b, n_f)
+        :param y: conditioning vector;   shape: (n_b, 1, n_f) or (n_b, n_f)
         :param sample_step: number of sampling steps
         :param guidance_strength: strength of conditional generation. It is not used if y is null.
         :param token_mask: token mask assigning unwanted token(s) with `True`;
-                                            shape: (1, 1, n_vocab)
+                                         shape: (1, 1, n_vocab)
         :param temperature: sampling temperature
         :type batch_size: int
         :type sequence_size: int
@@ -748,29 +809,14 @@ class ChemBFN(nn.Module):
         :type guidance_strength: float
         :type token_mask: torch.Tensor | None
         :type temperature: float
-        :return: sampled token indices;     shape: (n_b, n_t) \n
-                 entropy of the tokens;     shape: (n_b)
+        :return: sampled token indices;  shape: (n_b, n_t) \n
+                 entropy of the tokens;  shape: (n_b)
         :rtype: tuple
         """
         z = torch.zeros((batch_size, sequence_size, self.K), device=self.beta.device)
-        if y is not None:
-            y = self.reshape_y(y)
-        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
-            t = (i - 1).view(1, 1, 1).repeat(batch_size, 1, 1) / sample_step
-            theta = torch.softmax(z, -1)
-            beta = self.calc_beta(t + 1 / sample_step)
-            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
-            if token_mask is not None:
-                p = p.masked_fill_(token_mask, 0.0)
-            u = torch.randn_like(z)
-            z = (self.K * p - 1) * beta + (self.K * beta * temperature).sqrt() * u
-        t_final = torch.ones((batch_size, 1, 1), device=self.beta.device)
-        theta = torch.softmax(z, -1)
-        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
-        entropy = -(p * p.log()).sum(-1).mean(-1)
-        if token_mask is not None:
-            p = p.masked_fill_(token_mask, 0.0)
-        return torch.argmax(p, -1), entropy
+        return self._ode_process(
+            z, None, y, sample_step, guidance_strength, token_mask, temperature
+        )
 
     @torch.jit.export
     def inpaint(
@@ -800,30 +846,12 @@ class ChemBFN(nn.Module):
         :rtype: tuple
         """
         n_b, n_t = x.shape
-        mask = (x != 0).float()[..., None]
+        x_mask = (x != 0).float()[..., None]
         theta = torch.ones((n_b, n_t, self.K), device=x.device) / self.K
-        x_onehot = nn.functional.one_hot(x, self.K) * mask
-        theta = x_onehot + (1 - mask) * theta
-        if y is not None:
-            y = self.reshape_y(y)
-        for i in torch.linspace(1, sample_step, sample_step, device=x.device):
-            t = (i - 1).view(1, 1, 1).repeat(n_b, 1, 1) / sample_step
-            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
-            if token_mask is not None:
-                p = p.masked_fill_(token_mask, 0.0)
-            alpha = self.calc_discrete_alpha(t, t + 1 / sample_step)
-            e_k = nn.functional.one_hot(torch.argmax(p, -1), self.K).float()
-            mu = alpha * (self.K * e_k - 1)
-            sigma = (alpha * self.K).sqrt()
-            theta = (mu + sigma * torch.randn_like(mu)).exp() * theta
-            theta = theta / theta.sum(-1, True)
-            theta = x_onehot + (1 - mask) * theta
-        t_final = torch.ones((n_b, 1, 1), device=x.device)
-        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
-        entropy = -(p * p.log()).sum(-1).mean(-1)
-        if token_mask is not None:
-            p = p.masked_fill_(token_mask, 0.0)
-        return torch.argmax(p, -1), entropy
+        x_onehot = nn.functional.one_hot(x, self.K) * x_mask
+        theta = x_onehot + (1 - x_mask) * theta
+        mask = (x_onehot, x_mask)
+        return self._process(theta, mask, y, sample_step, guidance_strength, token_mask)
 
     @torch.jit.export
     def ode_inpaint(
@@ -856,29 +884,80 @@ class ChemBFN(nn.Module):
         :rtype: tuple
         """
         n_b, n_t = x.shape
-        mask = (x != 0).float()[..., None]
-        x_onehot = nn.functional.one_hot(x, self.K) * mask
+        x_mask = (x != 0).float()[..., None]
+        x_onehot = nn.functional.one_hot(x, self.K) * x_mask
         z = torch.zeros((n_b, n_t, self.K), device=self.beta.device)
-        if y is not None:
-            y = self.reshape_y(y)
-        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
-            t = (i - 1).view(1, 1, 1).repeat(n_b, 1, 1) / sample_step
-            theta = torch.softmax(z, -1)
-            theta = x_onehot + (1 - mask) * theta
-            beta = self.calc_beta(t + 1 / sample_step)
-            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
-            if token_mask is not None:
-                p = p.masked_fill_(token_mask, 0.0)
-            u = torch.randn_like(z)
-            z = (self.K * p - 1) * beta + (self.K * beta * temperature).sqrt() * u
-        t_final = torch.ones((n_b, 1, 1), device=self.beta.device)
-        theta = torch.softmax(z, -1)
-        theta = x_onehot + (1 - mask) * theta
-        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
-        entropy = -(p * p.log()).sum(-1).mean(-1)
-        if token_mask is not None:
-            p = p.masked_fill_(token_mask, 0.0)
-        return torch.argmax(p, -1), entropy
+        mask = (x_onehot, x_mask)
+        return self._ode_process(
+            z, mask, y, sample_step, guidance_strength, token_mask, temperature
+        )
+
+    @torch.jit.export
+    def optimise(
+        self,
+        x: Tensor,
+        y: Optional[Tensor] = None,
+        sample_step: int = 100,
+        guidance_strength: float = 4.0,
+        token_mask: Optional[Tensor] = None,
+    ) -> Tuple[Tensor, Tensor]:
+        """
+        Optimise the template molecule (mol2mol). \n
+        This method is equivalent to sampling from a customised prior distribution.
+
+        :param x: categorical indices of template;  shape: (n_b, n_t)
+        :param y: conditioning vector;              shape: (n_b, 1, n_f) or (n_b, n_f)
+        :param sample_step: number of sampling steps
+        :param guidance_strength: strength of conditional generation. It is not used if y is null.
+        :param token_mask: token mask assigning unwanted token(s) with `True`;
+                                                    shape: (1, 1, n_vocab)
+        :type x: torch.Tensor
+        :type y: torch.Tensor | None
+        :type sample_step: int
+        :type guidance_strength: float
+        :type token_mask: torch.Tensor | None
+        :return: sampled token indices;             shape: (n_b, n_t) \n
+                 entropy of the tokens;             shape: (n_b)
+        :rtype: tuple
+        """
+        x_onehot = nn.functional.one_hot(x, self.K).float()
+        theta = softmax(x_onehot, -1)
+        return self._process(theta, None, y, sample_step, guidance_strength, token_mask)
+
+    @torch.jit.export
+    def ode_optimise(
+        self,
+        x: Tensor,
+        y: Optional[Tensor] = None,
+        sample_step: int = 100,
+        guidance_strength: float = 4.0,
+        token_mask: Optional[Tensor] = None,
+        temperature: float = 0.5,
+    ) -> Tuple[Tensor, Tensor]:
+        """
+        ODE mol2mol.
+
+        :param x: categorical indices of template;  shape: (n_b, n_t)
+        :param y: conditioning vector;              shape: (n_b, 1, n_f) or (n_b, n_f)
+        :param sample_step: number of sampling steps
+        :param guidance_strength: strength of conditional generation. It is not used if y is null.
+        :param token_mask: token mask assigning unwanted token(s) with `True`;
+                                                    shape: (1, 1, n_vocab)
+        :param temperature: sampling temperature
+        :type x: torch.Tensor
+        :type y: torch.Tensor | None
+        :type sample_step: int
+        :type guidance_strength: float
+        :type token_mask: torch.Tensor | None
+        :type temperature: float
+        :return: sampled token indices;             shape: (n_b, n_t) \n
+                 entropy of the tokens;             shape: (n_b)
+        :rtype: tuple
+        """
+        z = nn.functional.one_hot(x, self.K).float()
+        return self._ode_process(
+            z, None, y, sample_step, guidance_strength, token_mask, temperature
+        )
 
     def inference(
         self, x: Tensor, mlp: MLP, embed_fn: Optional[Callable[[Tensor], Tensor]] = None
@@ -1052,22 +1131,6 @@ class EnsembleChemBFN(ChemBFN):
                     module.lora_dropout = None
             v.lora_enabled = False
 
-    def construct_y(
-        self, c: Union[List[Tensor], Dict[str, Tensor]]
-    ) -> Dict[str, Tensor]:
-        assert (
-            isinstance(c, dict) is self._label_is_dict
-        ), f"`c` should be a {'`dict` instance' if self._label_is_dict else '`list` instance'} but got {type(c)} instand."
-        out: Dict[str, Tensor] = {}
-        if isinstance(c, list):
-            c = dict(zip([f"val_{i}" for i in range(len(c))], c))
-        for name, model in self.cond_heads.items():
-            y = model.forward(c[name])
-            if y.dim() == 2:
-                y = y[:, None, :]
-            out[name] = y
-        return out
-
     def discrete_output_distribution(
         self, theta: Tensor, t: Tensor, y: Dict[str, Tensor], w: float
     ) -> Tensor:
@@ -1102,8 +1165,27 @@ class EnsembleChemBFN(ChemBFN):
             p_cond += p_cond_ * self.adapter_weights[name]
         return softmax((1 + w) * p_cond - w * p_uncond, -1)
 
+    def _map_to_dict(
+        self, c: Union[List[Tensor], Dict[str, Tensor]]
+    ) -> Dict[str, Tensor]:
+        assert (
+            isinstance(c, dict) is self._label_is_dict
+        ), f"`c` should be a {'`dict` instance' if self._label_is_dict else '`list` instance'} but got {type(c)} instand."
+        assert len(c) == len(
+            self.models
+        ), f"Number of conditions should match the number of LoRA models. We have {len(self.models)} LoRA models but {len(c)} conditions were provided."
+        out: Dict[str, Tensor] = {}
+        if isinstance(c, list):
+            c = dict(zip([f"val_{i}" for i in range(len(c))], c))
+        for name, model in self.cond_heads.items():
+            y = model.forward(c[name])
+            if y.dim() == 2:
+                y = y[:, None, :]
+            out[name] = y
+        return out
+
     @staticmethod
-    def reshape_y(y: Dict[str, Tensor]) -> Dict[str, Tensor]:
+    def _reshape(y: Dict[str, Tensor]) -> Dict[str, Tensor]:
         for k in y:
             assert y[k].dim() <= 3
             if y[k].dim() == 2:
@@ -1139,7 +1221,7 @@ class EnsembleChemBFN(ChemBFN):
                  entropy of the tokens;          shape: (n_b)
         :rtype: tuple
         """
-        y = self.construct_y(conditions)
+        y = self._map_to_dict(conditions)
         return super().sample(
             batch_size, sequence_size, y, sample_step, guidance_strength, token_mask
         )
@@ -1176,7 +1258,7 @@ class EnsembleChemBFN(ChemBFN):
                  entropy of the tokens;          shape: (n_b)
         :rtype: tuple
         """
-        y = self.construct_y(conditions)
+        y = self._map_to_dict(conditions)
         return super().ode_sample(
             batch_size,
             sequence_size,
@@ -1213,7 +1295,7 @@ class EnsembleChemBFN(ChemBFN):
                  entropy of the tokens;             shape: (n_b)
         :rtype: tuple
         """
-        y = self.construct_y(conditions)
+        y = self._map_to_dict(conditions)
         return super().inpaint(x, y, sample_step, guidance_strength, token_mask)
 
     @torch.inference_mode()
@@ -1245,11 +1327,76 @@ class EnsembleChemBFN(ChemBFN):
                  entropy of the tokens;             shape: (n_b)
         :rtype: tuple
         """
-        y = self.construct_y(conditions)
+        y = self._map_to_dict(conditions)
         return super().ode_inpaint(
             x, y, sample_step, guidance_strength, token_mask, temperature
         )
 
+    @torch.inference_mode()
+    def optimise(
+        self,
+        x: Tensor,
+        conditions: Union[List[Tensor], Dict[str, Tensor]],
+        sample_step: int = 100,
+        guidance_strength: float = 4.0,
+        token_mask: Optional[Tensor] = None,
+    ) -> Tuple[Tensor, Tensor]:
+        """
+        Optimise the template molecule (mol2mol). \n
+        This method is equivalent to sampling from a customised prior distribution.
+
+        :param x: categorical indices of template;  shape: (n_b, n_t)
+        :param conditions: conditioning vector;     shape: (n_b, n_c) * n_h
+        :param sample_step: number of sampling steps
+        :param guidance_strength: strength of conditional generation. It is not used if y is null.
+        :param token_mask: token mask assigning unwanted token(s) with `True`;
+                                                    shape: (1, 1, n_vocab)
+        :type x: torch.Tensor
+        :type y: torch.Tensor | None
+        :type sample_step: int
+        :type guidance_strength: float
+        :type token_mask: torch.Tensor | None
+        :return: sampled token indices;             shape: (n_b, n_t) \n
+                 entropy of the tokens;             shape: (n_b)
+        :rtype: tuple
+        """
+        y = self._map_to_dict(conditions)
+        return super().optimise(x, y, sample_step, guidance_strength, token_mask)
+
+    @torch.inference_mode()
+    def ode_optimise(
+        self,
+        x: Tensor,
+        conditions: Union[List[Tensor], Dict[str, Tensor]],
+        sample_step: int = 100,
+        guidance_strength: float = 4.0,
+        token_mask: Optional[Tensor] = None,
+        temperature: float = 0.5,
+    ) -> Tuple[Tensor, Tensor]:
+        """
+        ODE inpainting.
+
+        :param x: categorical indices of template;  shape: (n_b, n_t)
+        :param conditions: conditioning vector;     shape: (n_b, n_c) * n_h
+        :param sample_step: number of sampling steps
+        :param guidance_strength: strength of conditional generation. It is not used if y is null.
+        :param token_mask: token mask;              shape: (1, 1, n_vocab)
+        :param temperature: sampling temperature
+        :type x: torch.Tensor
+        :type conditions: list | dict
+        :type sample_step: int
+        :type guidance_strength: float
+        :type token_mask: torch.Tensor | None
+        :type temperature: float
+        :return: sampled token indices;             shape: (n_b, n_t) \n
+                 entropy of the tokens;             shape: (n_b)
+        :rtype: tuple
+        """
+        y = self._map_to_dict(conditions)
+        return super().ode_optimise(
+            x, y, sample_step, guidance_strength, token_mask, temperature
+        )
+
     def quantise(
         self, quantise_method: Optional[Callable[[ChemBFN], None]] = None
     ) -> None:
