@@ -7,9 +7,8 @@ import csv
 import random
 import warnings
 from pathlib import Path
-from typing import List, Dict, Tuple, Union, Optional
+from typing import List, Dict, Tuple, Union, Optional, Literal
 import torch
-import colorama
 import numpy as np
 from torch import cuda, Tensor, softmax
 from torch.utils.data import DataLoader
@@ -24,15 +23,7 @@ from rdkit.Chem import (
     AddHs,
     Mol,
 )
-from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles  # type: ignore
-from sklearn.metrics import (
-    roc_auc_score,
-    auc,
-    precision_recall_curve,
-    r2_score,
-    mean_absolute_error,
-    root_mean_squared_error,
-)
+from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles
 from .data import VOCAB_KEYS
 from .model import ChemBFN, MLP, EnsembleChemBFN
 
@@ -45,12 +36,74 @@ def _find_device() -> torch.device:
     return torch.device("cpu")
 
 
+def _parse_and_assert_param(
+    model: Union[ChemBFN, EnsembleChemBFN],
+    y: Optional[Union[Tensor, Dict[str, Tensor], List[Tensor]]],
+    method: str,
+) -> Optional[float]:
+    assert method.split(":")[0].lower() in ("ode", "bfn")
+    if isinstance(model, EnsembleChemBFN):
+        assert y is not None, "conditioning is required while using an ensemble model."
+        assert isinstance(y, list) or isinstance(y, dict)
+    else:
+        assert isinstance(y, Tensor) or (y is None)
+    if "ode" in method.lower():
+        tp = float(method.split(":")[-1])
+        assert tp > 0, "Sampling temperature should be higher than 0."
+        return tp
+    return None
+
+
+def _map_to_device(
+    y: Optional[Union[Tensor, Dict[str, Tensor], List[Tensor]]],
+    device: Union[str, torch.device],
+) -> Optional[Union[Tensor, Dict[str, Tensor], List[Tensor]]]:
+    if y is not None:
+        if isinstance(y, Tensor):
+            y = y.to(device)
+        elif isinstance(y, list):
+            y = [i.to(device) for i in y]
+        elif isinstance(y, dict):
+            y = {k: v.to(device) for k, v in y.items()}
+        else:
+            raise NotImplementedError
+    return y
+
+
+def _build_token_mask(
+    allowed_tokens: Union[str, List[str]],
+    vocab_keys: List[str],
+    device: Union[str, torch.tensor],
+) -> Optional[Tensor]:
+    if isinstance(allowed_tokens, list):
+        token_mask = [0 if i in allowed_tokens else 1 for i in vocab_keys]
+        token_mask = torch.tensor([[token_mask]], dtype=torch.bool).to(device)
+    else:
+        token_mask = None
+    return token_mask
+
+
+def _token_to_seq(
+    tokens: Tensor, entropy: Tensor, vocab_keys: List[str], separator: str, sort: bool
+) -> List[str]:
+    if sort:
+        sorted_idx = entropy.argsort(stable=True)
+        tokens = tokens[sorted_idx]
+    return [
+        separator.join([vocab_keys[i] for i in j])
+        .split("<start>" + separator)[-1]
+        .split(separator + "<end>")[0]
+        .replace("<pad>", "")
+        for j in tokens
+    ]
+
+
 @torch.no_grad()
 def test(
     model: ChemBFN,
     mlp: MLP,
     data: DataLoader,
-    mode: str = "regression",
+    mode: Literal["regression", "classification"] = "regression",
     device: Union[str, torch.device, None] = None,
 ) -> Dict[str, float]:
     """
@@ -86,6 +139,12 @@ def test(
         predict_y.append(y_hat.detach().to("cpu"))
     predict_y, label_y = torch.cat(predict_y, 0), torch.cat(label_y, 0).split(1, -1)
     if mode == "regression":
+        from sklearn.metrics import (
+            r2_score,
+            mean_absolute_error,
+            root_mean_squared_error,
+        )
+
         predict_y = [
             predict[label_y[i] != torch.inf]
             for (i, predict) in enumerate(predict_y.split(1, -1))
@@ -99,6 +158,8 @@ def test(
         r2 = [r2_score(label, predict) for (label, predict) in y_zipped]
         return {"MAE": mae, "RMSE": rmse, "R^2": r2}
     if mode == "classification":
+        from sklearn.metrics import roc_auc_score, auc, precision_recall_curve
+
         n_c = len(label_y)
         predict_y = predict_y.chunk(n_c, -1)
         y_zipped = list(zip(label_y, predict_y))
@@ -123,7 +184,9 @@ def test(
 
 
 def split_dataset(
-    file: Union[str, Path], split_ratio: List[int] = [8, 1, 1], method: str = "random"
+    file: Union[str, Path],
+    split_ratio: List[int] = [8, 1, 1],
+    method: Literal["random", "scaffold"] = "random",
 ) -> None:
     """
     Split a dataset.
@@ -142,7 +205,6 @@ def split_dataset(
     assert file.endswith(".csv")
     assert len(split_ratio) == 3
     assert method in ("random", "scaffold")
-    colorama.just_fix_windows_console()
     with open(file, "r") as f:
         data = list(csv.reader(f))
     header = data[0]
@@ -167,10 +229,8 @@ def split_dataset(
             # compute Bemis-Murcko scaffold
             if len(smiles_idx) > 1:
                 warnings.warn(
-                    "\033[32;1m"
                     f"We found {len(smiles_idx)} SMILES strings in a row!"
-                    " Only the first SMILES will be used to compute the molecular scaffold."
-                    "\033[0m",
+                    " Only the first SMILES will be used to compute the molecular scaffold.",
                     stacklevel=2,
                 )
             try:
@@ -197,10 +257,10 @@ def split_dataset(
     with open(file.replace(".csv", "_test.csv"), "w", newline="") as fte:
         writer = csv.writer(fte)
         writer.writerows([header] + test_set)
-    with open(file.replace(".csv", "_val.csv"), "w", newline="") as fva:
-        writer = csv.writer(fva)
-        writer.writerows([header] + val_set)
-    colorama.deinit()
+    if val_set:
+        with open(file.replace(".csv", "_val.csv"), "w", newline="") as fva:
+            writer = csv.writer(fva)
+            writer.writerows([header] + val_set)
 
 
 @torch.no_grad()
@@ -219,7 +279,7 @@ def sample(
     sort: bool = False,
 ) -> List[str]:
     """
-    Sampling.
+    Sampling molecules.
 
     :param model: trained ChemBFN model
     :param batch_size: batch size
@@ -250,32 +310,12 @@ def sample(
     :return: a list of generated molecular strings
     :rtype: list
     """
-    assert method.split(":")[0].lower() in ("ode", "bfn")
-    if isinstance(model, EnsembleChemBFN):
-        assert y is not None, "conditioning is required while using an ensemble model."
-        assert isinstance(y, list) or isinstance(y, dict)
-    else:
-        assert isinstance(y, Tensor) or y is None
-    if device is None:
-        device = _find_device()
+    tp = _parse_and_assert_param(model, y, method)
+    device = _find_device() if device is None else device
     model.to(device).eval()
-    if y is not None:
-        if isinstance(y, Tensor):
-            y = y.to(device)
-        elif isinstance(y, list):
-            y = [i.to(device) for i in y]
-        elif isinstance(y, dict):
-            y = {k: v.to(device) for k, v in y.items()}
-        else:
-            raise NotImplementedError
-    if isinstance(allowed_tokens, list):
-        token_mask = [0 if i in allowed_tokens else 1 for i in vocab_keys]
-        token_mask = torch.tensor([[token_mask]], dtype=torch.bool).to(device)
-    else:
-        token_mask = None
-    if "ode" in method.lower():
-        tp = float(method.split(":")[-1])
-        assert tp > 0, "Sampling temperature should be higher than 0."
+    y = _map_to_device(y, device)
+    token_mask = _build_token_mask(allowed_tokens, vocab_keys, device)
+    if tp:
         tokens, entropy = model.ode_sample(
             batch_size, sequence_size, y, sample_step, guidance_strength, token_mask, tp
         )
@@ -283,16 +323,7 @@ def sample(
         tokens, entropy = model.sample(
             batch_size, sequence_size, y, sample_step, guidance_strength, token_mask
         )
-    if sort:
-        sorted_idx = entropy.argsort(stable=True)
-        tokens = tokens[sorted_idx]
-    return [
-        seperator.join([vocab_keys[i] for i in j])
-        .split("<start>" + seperator)[-1]
-        .split(seperator + "<end>")[0]
-        .replace("<pad>", "")
-        for j in tokens
-    ]
+    return _token_to_seq(tokens, entropy, vocab_keys, seperator, sort)
 
 
 @torch.no_grad()
@@ -339,33 +370,13 @@ def inpaint(
     :return: a list of generated molecular strings
     :rtype: list
     """
-    assert method.split(":")[0].lower() in ("ode", "bfn")
-    if isinstance(model, EnsembleChemBFN):
-        assert y is not None, "conditioning is required while using an ensemble model."
-        assert isinstance(y, list) or isinstance(y, dict)
-    else:
-        assert isinstance(y, Tensor) or y is None
-    if device is None:
-        device = _find_device()
+    tp = _parse_and_assert_param(model, y, method)
+    device = _find_device() if device is None else device
     model.to(device).eval()
     x = x.to(device)
-    if y is not None:
-        if isinstance(y, Tensor):
-            y = y.to(device)
-        elif isinstance(y, list):
-            y = [i.to(device) for i in y]
-        elif isinstance(y, dict):
-            y = {k: v.to(device) for k, v in y.items()}
-        else:
-            raise NotImplementedError
-    if isinstance(allowed_tokens, list):
-        token_mask = [0 if i in allowed_tokens else 1 for i in vocab_keys]
-        token_mask = torch.tensor([[token_mask]], dtype=torch.bool).to(device)
-    else:
-        token_mask = None
-    if "ode" in method.lower():
-        tp = float(method.split(":")[-1])
-        assert tp > 0, "Sampling temperature should be higher than 0."
+    y = _map_to_device(y, device)
+    token_mask = _build_token_mask(allowed_tokens, vocab_keys, device)
+    if tp:
         tokens, entropy = model.ode_inpaint(
             x, y, sample_step, guidance_strength, token_mask, tp
         )
@@ -373,16 +384,68 @@ def inpaint(
         tokens, entropy = model.inpaint(
             x, y, sample_step, guidance_strength, token_mask
         )
-    if sort:
-        sorted_idx = entropy.argsort(stable=True)
-        tokens = tokens[sorted_idx]
-    return [
-        separator.join([vocab_keys[i] for i in j])
-        .split("<start>" + separator)[-1]
-        .split(separator + "<end>")[0]
-        .replace("<pad>", "")
-        for j in tokens
-    ]
+    return _token_to_seq(tokens, entropy, vocab_keys, separator, sort)
+
+
+@torch.no_grad()
+def optimise(
+    model: Union[ChemBFN, EnsembleChemBFN],
+    x: Tensor,
+    sample_step: int = 100,
+    y: Optional[Union[Tensor, Dict[str, Tensor], List[Tensor]]] = None,
+    guidance_strength: float = 4.0,
+    device: Union[str, torch.device, None] = None,
+    vocab_keys: List[str] = VOCAB_KEYS,
+    separator: str = "",
+    method: str = "BFN",
+    allowed_tokens: Union[str, List[str]] = "all",
+    sort: bool = False,
+) -> List[str]:
+    """
+    Optimising template molecules (mol2mol).
+
+    :param model: trained ChemBFN model
+    :param x: categorical indices of template;  shape: (n_b, n_t)
+    :param sample_step: number of sampling steps
+    :param y: conditioning vector;              shape: (n_b, 1, n_f) or (n_b, n_f) \n
+              or a list/`dict` of conditions;   shape: (n_b, n_c) * n_h
+
+    :param guidance_strength: strength of conditional generation. It is not used if y is null.
+    :param device: hardware accelerator
+    :param vocab_keys: a list of (ordered) vocabulary
+    :param separator: token separator; default is `""`
+    :param method: sampling method chosen from `"ODE:x"` or `"BFN"` where `x` is the value of sampling temperature; default is `"BFN"`
+    :param allowed_tokens: a list of allowed tokens; default is `"all"`
+    :param sort: whether to sort the samples according to entropy values; default is `False`
+    :type model: bayesianflow_for_chem.model.ChemBFN | bayesianflow_for_chem.model.EnsembleChemBFN
+    :type x: torch.Tensor
+    :type sample_step: int
+    :type y: torch.Tensor | list | dict | None
+    :type guidance_strength: float
+    :type device: str | torch.device | None
+    :type vocab_keys: list
+    :type separator: str
+    :type method: str
+    :type allowed_tokens: str | list
+    :type sort: bool
+    :return: a list of generated molecular strings
+    :rtype: list
+    """
+    tp = _parse_and_assert_param(model, y, method)
+    device = _find_device() if device is None else device
+    model.to(device).eval()
+    x = x.to(device)
+    y = _map_to_device(y, device)
+    token_mask = _build_token_mask(allowed_tokens, vocab_keys, device)
+    if tp:
+        tokens, entropy = model.ode_optimise(
+            x, y, sample_step, guidance_strength, token_mask, tp
+        )
+    else:
+        tokens, entropy = model.optimise(
+            x, y, sample_step, guidance_strength, token_mask
+        )
+    return _token_to_seq(tokens, entropy, vocab_keys, separator, sort)
 
 
 def quantise_model_(model: ChemBFN) -> None:
@@ -465,7 +528,7 @@ class GeometryConverter:
     def smiles2cartesian(
         smiles: str,
         num_conformers: int = 250,
-        rdkit_ff_type: str = "MMFF",
+        rdkit_ff_type: Literal["MMFF", "UFF"] = "MMFF",
         refine_with_crest: bool = False,
         spin: float = 0.0,
     ) -> Tuple[List[str], np.ndarray]:
